/*******************************************************************************
 viewer_plotting.js: client module for plot mode

 $Author: dmc $
 $Revision: 3083 $
 $Date: 2022-07-08 12:15:34 -0300 (Fri, 08 Jul 2022) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_plotting.js $
*******************************************************************************/

/* 
scope and hold settings will be applied to the next shot by the server

string.shot_params.scope --> shot.scope = true/false (whether it has been set)
string.shot_params.scope_x --> shot.scope_x = moa (windage)
string.shot_params.scope_y --> shot.scope_y = moa (elevation)
string.shot_params.hold --> shot.hold = true/false (whether it has been set)
string.shot_params.hold_x --> shot.hold_x = moa (windage)
string.shot_params.hold_y --> shot.hold_y = moa (elevation)
*/

plotting = new function() {

	this.init = function() {
		this.div_slider_x = $("#divPlotSliderX");
		this.div_slider_y = $("#divPlotSliderY");
	}

	this.clear = function() {
		hide(this.div_slider_x);
		this.div_slider_x.empty();

		hide(this.div_slider_y);
		this.div_slider_y.empty();
	}
	
	this.enable_view = false; // show sliders and shots on plot instead of impacts
	this.enable_controls = false; // allow scope and hold controls

	this.canvas_x = null;
	this.canvas_y = null;
	
	this.args_x = { pan_x: 0, pan_y: 0 };
	this.args_y = { pan_x: 0, pan_y: 0 };

	this.update = function() {
		this.clear();
		if (!cur_string()) return;
		if (!this.enable_view) return;

		//log("plotting.update()");
		
		var thickness = 70;

		set_pos(this.div_slider_y, "top", toolbar.div[0].offsetTop + toolbar.div[0].offsetHeight + 10);
		set_pos(this.div_slider_y, "bottom", targetrow.div[0].offsetHeight + thickness + 25);
		set_pos(this.div_slider_y, "width", thickness);
		set_pos(this.div_slider_y, "left", 10);

		set_pos(this.div_slider_x, "left", this.div_slider_y[0].offsetLeft + this.div_slider_y[0].offsetWidth + 10);
		set_pos(this.div_slider_x, "right", 10);
		set_pos(this.div_slider_x, "bottom", targetrow.div[0].offsetHeight + 15);
		set_pos(this.div_slider_x, "height", thickness);

		// create or resize each canvas
		this.canvas_x = this.build_canvas(this.div_slider_x, this.canvas_x, this.args_x, true);
		this.canvas_y = this.build_canvas(this.div_slider_y, this.canvas_y, this.args_y, false);

		// redraw both canvas
		this.update_canvas();

		show(this.div_slider_x);
		show(this.div_slider_y);
		//log("plotting.update() done");
	}
	
	this.build_canvas = function(div, canvas, args, wind) {
		
		args.w = div.width();
		args.h = div.height();
		args.wind = wind;

		// rebuild canvas if it does not already exist
		if (ok(canvas)) {

			if (canvas[0].width != args.w || canvas[0].height != args.h) {
				canvas[0].width = args.w;
				canvas[0].height = args.h;
			}

		} else {

			canvas = $("<canvas>").attr({width: args.w, height: args.h});
			var events = new Hammer(canvas[0]);
			
			canvas[0].events = events;
			canvas[0].args = args;

			events.get("pan").set({ direction: Hammer.DIRECTION_ALL });

			events.on("panstart", function(e) {
				var args = e.target.args;
				
				args.pan_last_x = 0;
				args.pan_last_y = 0;
				
				// initialize string scope settings if not yet done
				var params = cur_string().data.shot_params;
				if (!params.scope) {
					params.scope = true;
					params.scope_x = [0, 0, 0]; // one value for each shooter
					params.scope_y = [0, 0, 0];
					
					// need to init server side as well
					if (plotting.enable_controls) { // not in view-only mode
						cur_string().crud({ shot_params: {
							scope: true,
							scope_x: [0, 0, 0],
							scope_y: [0, 0, 0]
						} });
					}
				}
				
				// args for display contains only the client-side value for the active shooter
				args.pan_scope = true;
				args.pan_scope_x = params.scope_x[cur_shooter];
				args.pan_scope_y = params.scope_y[cur_shooter];
			});
			
			// adjust scope setting continuously client-side
			events.on("pan", function(e) {
				var args = e.target.args;
				
				var dx = e.deltaX - args.pan_last_x; args.pan_last_x = e.deltaX;
				var dy = e.deltaY - args.pan_last_y; args.pan_last_y = e.deltaY;

				if (args.wind) {
					args.pan_scope_x += dx / args.moa_to_px;
				} else {
					args.pan_scope_y -= dy / args.moa_to_px;
				}

				maintarget.update_canvas();
				plotting.update_canvas();
			});
			
			// send to server when adjustment is complete
			events.on("panend", function(e) {
				var args = e.target.args;

				if (plotting.enable_controls) { // not in view-only mode
				
					// use local partial array to only crud cur_shooter's data
					// to avoid conflicts with other shooters
					var new_scope_x = [];
					var new_scope_y = [];
					if (g_config.option_grid_unit == "mil" || g_config.option_grid_unit == "tenthmil") {
						new_scope_x[cur_shooter] = mil_to_moa(Math.round(moa_to_mil(args.pan_scope_x) * 10) / 10);
						new_scope_y[cur_shooter] = mil_to_moa(Math.round(moa_to_mil(args.pan_scope_y) * 10) / 10);
					} else { // moa
						new_scope_x[cur_shooter] = Math.round(args.pan_scope_x * 8) / 8;
						new_scope_y[cur_shooter] = Math.round(args.pan_scope_y * 8) / 8;
					}
					
					log("scope set for shooter " + cur_shooter + ": " + new_scope_x[cur_shooter] + ", " + new_scope_y[cur_shooter]);
					
					var crud = { shot_params: {
									scope: true,
									scope_x: new_scope_x,
									scope_y: new_scope_y
								} };
								
					update_object(cur_string().data, crud); // update locally for instant feedback
					cur_string().crud(crud); // send to server
				}
				
				args.pan_scope = false;

				maintarget.update_canvas();
				plotting.update_canvas();
			});

		}
		
		div.append(canvas);

		return canvas;
	}
	
	this.update_canvas = function() {
		this.update_canvas_each(this.canvas_x, true);
		this.update_canvas_each(this.canvas_y, false);
	}
	
	this.update_canvas_each = function(canvas, wind) {
		if (!ok(canvas)) return;
		if (!ok(canvas[0].parentElement)) return;
		
		var use_mils = (g_config.option_grid_unit == "mil" || g_config.option_grid_unit == "tenthmil");

		var g = canvas[0].getContext("2d");
		
		var args = (wind)?(this.args_x):(this.args_y);
		args.moa_to_px = maintarget.args.moa_to_px;
		args.mm_to_px = maintarget.args.mm_to_px;
		
		var params = cur_string().data.shot_params;

		//g.fillStyle = "#555";
		g.clearRect(0, 0, args.w, args.h);
		
		var sx = args.w, sy = 0; // scope setting

		if (wind) {
			args.cx = maintarget.args.cx + maintarget.div[0].offsetLeft - canvas[0].offsetLeft - canvas[0].parentElement.offsetLeft + 2;
			args.cy = 0;
		} else {
			args.cy = maintarget.args.cy + maintarget.div[0].offsetTop - canvas[0].offsetTop - canvas[0].parentElement.offsetTop - 2;
			args.cx = args.w;
		}
		
		// current scope setting (per shooter)
		var scope_x = 0, scope_y = 0;
		if (args.pan_scope) {
			scope_x = args.pan_scope_x;
			scope_y = args.pan_scope_y;
		} else if (params.scope) {
			if (typeof(params.scope_x) == "object") {
				scope_x = params.scope_x[cur_shooter];
				scope_y = params.scope_y[cur_shooter];
			} else {
				scope_x = params.scope_x;
				scope_y = params.scope_y;
			}
		}
		
		// current hold setting (global)
		var hold_x = 0, hold_y = 0;
		if (maintarget.args.pan_hold) {
			hold_x = maintarget.args.pan_hold_x;
			hold_y = maintarget.args.pan_hold_y;
		} else if (params.hold) {
			hold_x = params.hold_x;
			hold_y = params.hold_y;
		}

		// gridlines
		var stroke_gridlines;
		if (wind) {

			stroke_gridlines = function(y1, y2, increment) {
				var x_left, x_right;
				if (use_mils) {
					x_left = mil_to_moa(Math.floor(moa_to_mil(-scope_x + (0 - args.cx) / args.moa_to_px)));
					x_right = mil_to_moa(Math.ceil(moa_to_mil(-scope_x + (args.w - args.cx) / args.moa_to_px)));
				} else {
					x_left = Math.floor(-scope_x + (0 - args.cx) / args.moa_to_px);
					x_right = Math.ceil(-scope_x + (args.w - args.cx) / args.moa_to_px);
				}
				for (var x_moa = x_left; x_moa <= x_right; x_moa += increment) {
					var x = args.cx + round((scope_x + x_moa) * args.moa_to_px);
					var scale = Math.cos(Math.PI * 0.9 * (x - args.cx) / args.w);
					if (scale > 0) {
						g.beginPath();
						g.strokeStyle = "rgba(200, 200, 200, " + scale + ")";
						g.moveTo(round(x), round(args.h - y1));
						g.lineTo(round(x), round(args.h - y1 - (y2 - y1) * scale));
						g.stroke();
					}
				}
			}
			
		} else {
			
			stroke_gridlines = function(x1, x2, increment) {
				var y_top, y_bottom;
				if (use_mils) {
					y_top = mil_to_moa(Math.ceil(moa_to_mil(-scope_y + (args.cy - 0) / args.moa_to_px)));
					y_bottom = mil_to_moa(Math.floor(moa_to_mil(-scope_y + (args.cy - args.h) / args.moa_to_px)));
				} else {
					y_top = Math.ceil(-scope_y + (args.cy - 0) / args.moa_to_px);
					y_bottom = Math.floor(-scope_y + (args.cy - args.h) / args.moa_to_px);
				}
				for (var y_moa = y_bottom; y_moa <= y_top; y_moa += increment) {
					var y = args.cy - round((scope_y + y_moa) * args.moa_to_px);
					var scale = Math.cos(Math.PI * 0.9 * (y - args.cy) / args.h);
					if (scale > 0) {
						g.beginPath();
						g.strokeStyle = "rgba(200, 200, 200, " + scale + ")";
						g.moveTo(x1, round(y));
						g.lineTo(round(x1 + (x2 - x1) * scale), round(y));
						g.stroke();
					}
				}
			}
			
		}
		
		// 1/4 moa or 1/10 mil (thin lines)
		g.lineWidth = 2;
		stroke_gridlines(25, 45, (use_mils)?(mil_to_moa(0.1)):(0.25));
		
		// 1 moa or 1 mil (thick lines)
		g.lineWidth = 6;
		stroke_gridlines(25, 50, (use_mils)?(mil_to_moa(0.5)):(1));
		
		// zero marker
		g.beginPath();
		if (wind) {
			g.moveTo(args.cx, args.h * 0.25);
			g.lineTo(args.cx - 15, 0);
			g.lineTo(args.cx + 15, 0);
		} else {
			g.moveTo(args.w * (1 - 0.25), args.cy);
			g.lineTo(args.w, args.cy - 15);
			g.lineTo(args.w, args.cy + 15);
		}
		g.fillStyle = "#FF4";
		g.fill();

		// shots
		var shots = cur_string().shots();
		
		// only show current shooter's shots
		var shots_filtered = [];
		for (var i in shots) {
			var shot = shots[i];
			if (!shot || !shot.valid || shot.strikeout) continue;
			if (cur_string().data.shooter_count > 1 && shot.shooter_num != cur_shooter) continue;
			shots_filtered.push(shot);
		}
		shots = shots_filtered;




		// draw windage and elevation graph with "dots".
		var n_shots = 25;
		for (var i = (shots.length > n_shots)?(shots.length - n_shots):(0); i < shots.length; i++) {
			var shot = shots[i];
			var hideByDelay = false;

            if (isShotHiddenByDelay(shot, cur_string())) continue; // 23-Nov-2020 fix "ZA Cheat Mode"
			
			var count = shots.length - 1 - i;
			
			var adj_x = scope_x + hold_x;
			var adj_y = scope_y + hold_y;
			if (shot.scope) {
				if (typeof(shot.scope_x) == "object") {
					adj_x -= shot.scope_x[cur_shooter];
					adj_y -= shot.scope_y[cur_shooter];
				} else {
					adj_x -= shot.scope_x;
					adj_y -= shot.scope_y;
				}
			}
			if (shot.hold) {
				adj_x -= shot.hold_x;
				adj_y -= shot.hold_y;
			}
			
			var px, py;
			if (wind) {
				px = args.cx + (adj_x * args.moa_to_px + shot.x * args.mm_to_px);
				py = args.cy + (2 + count * 3);
				if (px < 0) px = 0;
				if (px > args.w) px = args.w;
			} else {
				py = args.cy - (adj_y * args.moa_to_px + shot.y * args.mm_to_px);
				px = args.cx - (2 + count * 3);
				if (py < 0) py = 0;
				if (py > args.h) py = args.h;
			}
			

			g.beginPath();
			var radius;
			if (shot == cur_string().cur_shot()) {
				var radius = 8;
				g.fillStyle = "#F88";
			} else {
				var radius = 5;
				g.fillStyle = "#FFF";
			}
			g.arc(px, py, radius, 0, 2*Math.PI);
			g.strokeStyle = "#000"
			g.lineWidth = 1;
			g.fill();
			g.stroke();

		}
		
		// scope text markers
		g.fillStyle = "#FFF";
		g.font = "18px Arial";

		if (wind) {

			g.textAlign = "center";
			g.textBaseline = "bottom";
			
			var x_left, x_right;
			if (use_mils) {
				x_left = mil_to_moa(Math.floor(moa_to_mil(-scope_x + (0 - args.cx) / args.moa_to_px)));
				x_right = mil_to_moa(Math.ceil(moa_to_mil(-scope_x + (args.w - args.cx) / args.moa_to_px)));
			} else {
				x_left = Math.floor(-scope_x + (0 - args.cx) / args.moa_to_px);
				x_right = Math.ceil(-scope_x + (args.w - args.cx) / args.moa_to_px);
			}
			
			var inc = (use_mils)?(mil_to_moa(0.5)):(1);

			for (var x_moa = x_left; x_moa <= x_right; x_moa += inc) {
				var x = args.cx + round((scope_x + x_moa) * args.moa_to_px);
				var scale = Math.cos(Math.PI * 0.9 * (x - args.cx) / args.w);
				if (scale > 0) {
					
					g.clearRect(-12 + round(x), -17 + round(args.h + 5 * (1 - scale)), 24, 30);

					g.beginPath();
					var moa_rounded = (use_mils)?(moa_to_mil(-x_moa).toFixed(1)):(round(-x_moa));
					var str = "0";
					if (moa_rounded < 0) str = (-moa_rounded) + "L";
					else if (moa_rounded > 0) str = moa_rounded + "R";
					g.fillStyle = "rgba(255, 255, 255, " + scale + ")";
					g.fillText(str, round(x), -3 + round(args.h + 8 * (1 - scale)));
				}
			}

		} else {

			g.textAlign = "center";
			g.textBaseline = "middle";
			
			var y_top, y_bottom;
			if (use_mils) {
				y_top = mil_to_moa(Math.ceil(moa_to_mil(-scope_y + (args.cy - 0) / args.moa_to_px)));
				y_bottom = mil_to_moa(Math.floor(moa_to_mil(-scope_y + (args.cy - args.h) / args.moa_to_px)));
			} else {
				y_top = Math.ceil(-scope_y + (args.cy - 0) / args.moa_to_px);
				y_bottom = Math.floor(-scope_y + (args.cy - args.h) / args.moa_to_px);
			}

			var inc = (use_mils)?(mil_to_moa(1)):(1);

			for (var y_moa = y_bottom; y_moa <= y_top; y_moa += inc) {
				var y = args.cy - round((scope_y + y_moa) * args.moa_to_px);
				var scale = Math.cos(Math.PI * 0.9 * (y - args.cy) / args.h);
				if (scale > 0) {
					
					g.clearRect(-12 + round(0 - 5 * (1 - scale)), -12 + round(y), 30, 24);

					g.fillStyle = "rgba(255, 255, 255, " + scale + ")";
					var str = (use_mils)?(round(moa_to_mil(-y_moa))):(round(-y_moa));
					if (str > 0) str = "+" + str;
					g.fillText(str, round(12 - 5 * (1 - scale)), round(y));
				}
			}
		}
		
	}

}


